from enum import Enum


class ProblemType(str, Enum):
    def __str__(self) -> str:
        return self.value

    # type check
    FLOAT = "a Float"
    STRING = "a String"
    NUMERIC = "numeric"
    BOOLEAN = "a Boolean"
    OBJECT = "an Object"
    ARRAY = "an Array"
    INTEGER = "an Integer"
    # pattern check
    URI = "A URI"
    REFERENCE_PATH = "a Reference Path"
    REFERENCE_PATH_OR_FUNC = "a Reference Path or an Intrinsic Function"
    JSON_PATH = "a JSONPath"
    JSONata = "a JSONata"

    UNKNOWN = "unknown"



from statelint.linter import Linter


def test_end():
    state_machine = {"States": {"x": {"Type": "Succeed", "End": True}}, "StartAt": "x"}
    assert Linter.validate(state_machine) == [
        'Field "End" not allowed in State Machine.States.x'
    ]


def test_fail_state_fields():
    state_machine = {
        "States": {
            "x": {
                "Type": "Fail",
                "Cause": "c",
                "Error": "e",
                "InputPath": "$",
                "OutputPath": "$",
                "Next": "y",
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x has forbidden field "InputPath"',
        'State Machine.States.x has forbidden field "OutputPath"',
        'State Machine.States.x has forbidden field "Next"',
        'No state found named "y", referenced at State Machine.States.x.Next',
    ]


def test_succeed_state_fields():
    state_machine = {
        "States": {
            "x": {
                "Type": "Succeed",
                "Cause": "c",
                "Error": "e",
                "InputPath": "$",
                "OutputPath": "$",
                "Next": "y",
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x has forbidden field "Next"',
        'Field "Cause" not allowed in State Machine.States.x',
        'Field "Error" not allowed in State Machine.States.x',
        'No state found named "y", referenced at State Machine.States.x.Next',
    ]



import pytest

from statelint.utils.jsonata_helper import evaluate_jsonata


@pytest.mark.parametrize(
    "expr,variables,expected",
    [
        ("$value", {"value": "42"}, "42"),
        ("$value", {"value": 42}, 42),
        ("$number($value)", {"value": "42"}, 42),
        ("$number($value)", {"value": 42}, 42),
        ("$number($value)", {"value": "A"}, None),
        ("$value(42)", {"value": 42}, None),
    ],
)
def test_evaluate_jsonata(expr, variables, expected):
    result = evaluate_jsonata(expr, variables)
    assert result == expected



from statelint.config import Config
from statelint.nodes.factory.factory import StateFactory
from statelint.nodes.map_state import MapState
from statelint.nodes.state_machine import StateMachine


def test_no_variable():
    sm = StateMachine(
        StateFactory(), {"States": {"x": {"Type": "Succeed"}}, "StartAt": "x"}, Config()
    )
    assert sm.validate() == []
    assert sm.variable_scopes == [{}]


def test_child_variable():
    variables = {"v1": 42, "v2": "str", "v3": {"v4": None}}
    sm = StateMachine(
        StateFactory(),
        {
            "QueryLanguage": "JSONata",
            "States": {
                "x": {"Type": "Pass", "Assign": variables, "Next": "y"},
                "y": {"Type": "Succeed"},
            },
            "StartAt": "x",
        },
        Config(),
    )
    assert sm.validate() == []
    assert sm.variable_scopes == [variables]


def test_nest():
    variables_in_parent = {"v1": 42}
    variables_in_child = {"v2": True}
    sm = StateMachine(
        StateFactory(),
        {
            "QueryLanguage": "JSONata",
            "States": {
                "x": {"Type": "Pass", "Assign": variables_in_parent, "Next": "y"},
                "y": {
                    "Type": "Map",
                    "Iterator": {
                        "StartAt": "y1",
                        "States": {
                            "y1": {
                                "Type": "Pass",
                                "Assign": variables_in_child,
                                "Next": "y2",
                            },
                            "y2": {"Type": "Fail"},
                        },
                    },
                    "End": True,
                },
            },
            "StartAt": "x",
        },
        Config(),
    )
    assert sm.validate() == []
    assert sm.variable_scopes == [variables_in_parent]
    map_state = sm._states["y"]
    assert isinstance(map_state, MapState)
    assert map_state.variable_scopes == [variables_in_parent]
    assert map_state._iterator
    assert map_state._iterator.variable_scopes == [
        variables_in_parent,
        variables_in_child,
    ]



import pytest

from statelint.linter import Linter


def test_null_type():
    state_machine = {"StartAt": "x", "States": {"x": {"Type": None, "End": True}}}
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Type should be non-null",
        'State Machine.States.x.Type is "", not one of the allowed values '
        '["Pass", "Succeed", "Fail", "Task", "Choice", "Wait", "Parallel", "Map"]',
        'Field "End" not allowed in State Machine.States.x',
    ]


def test_illegal_type():
    state_machine = {"StartAt": "x", "States": {"x": {"Type": "?", "End": True}}}
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.Type is "?", not one of the allowed values '
        '["Pass", "Succeed", "Fail", "Task", "Choice", "Wait", "Parallel", "Map"]',
        'Field "End" not allowed in State Machine.States.x',
    ]


def test_null_resource():
    state_machine = {
        "StartAt": "x",
        "States": {"x": {"Type": "Task", "Resource": 1, "End": True}},
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Resource is 1 but should be A URI"
    ]


@pytest.mark.parametrize(
    "seconds,problems",
    [
        (0, []),
        (-1, ["State Machine.States.x.Seconds is -1 but allowed floor is 0"]),
        (1, []),
    ],
)
def test_zero_seconds(seconds, problems):
    state_machine = {
        "StartAt": "x",
        "States": {"x": {"Type": "Wait", "Seconds": seconds, "End": True}},
    }
    assert Linter.validate(state_machine) == problems


@pytest.mark.parametrize("value,expected", [(42, 42), ("xyz", '"xyz"')])
def test_invalid_jsonata(value, expected):
    state_machine = {
        "StartAt": "x",
        "States": {
            "x": {
                "Type": "Task",
                "QueryLanguage": "JSONata",
                "Arguments": value,
                "End": True,
            }
        },
    }
    assert Linter.validate(state_machine) == [
        f"State Machine.States.x.Arguments is {expected} but should be a JSONata"
    ]


def test_unknown_query_language():
    state_machine = {
        "StartAt": "x",
        "QueryLanguage": "Yaml",
        "States": {
            "x": {
                "Type": "Fail",
            }
        },
    }
    assert Linter.validate(state_machine) == [
        'State Machine.QueryLanguage is "Yaml", not one of the allowed values '
        '["JSONPath", "JSONata"]',
    ]



from typing import Callable

from ..problem import ProblemPredicate, ProblemType
from ..utils.jsonata_helper import is_jsonata_value
from .base import Field
from .common import to_json
from .field_value import FieldValue
from .str_field import StrField


class JSONataField(Field):

    def __init__(self, name: str, acceptable_field: Callable[[str], Field]) -> None:
        super().__init__(name)
        self._field = acceptable_field(name)
        self._str_field = StrField(name)

    def validate(self, field_value: FieldValue) -> list[ProblemPredicate]:
        if not self._field.validate(field_value):
            return []
        value = field_value.value
        if not self._str_field.validate(field_value):
            if is_jsonata_value(value):
                return []

        problem_type = ProblemType.JSONata
        return [
            ProblemPredicate(
                f" is {to_json(value)} but should be {problem_type}", problem_type
            )
        ]



import pytest

from statelint.linter import Linter


@pytest.mark.parametrize(
    "error,cause",
    [
        ({}, {}),
        ({"Error": "error"}, {"Cause": "cause"}),
        ({"ErrorPath": "$.x"}, {"CausePath": "$.x"}),
        (
            {"ErrorPath": "States.Format('{} {}', 1, 2)"},
            {"CausePath": "States.Format('{} {}', 1, 2)"},
        ),
    ],
)
def test_ok(error, cause):
    state_machine = {
        "States": {"x": {"Type": "Fail", **error, **cause}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


@pytest.mark.parametrize(
    "props,errors",
    [
        (
            {"Error": 42, "Cause": 42},
            [
                "State Machine.States.x.Error is 42 but should be a String",
                "State Machine.States.x.Cause is 42 but should be a String",
            ],
        ),
        (
            {"ErrorPath": "x", "CausePath": "x"},
            [
                'State Machine.States.x.ErrorPath is "x" but should be a Reference '
                "Path or an Intrinsic Function",
                'State Machine.States.x.CausePath is "x" but should be a Reference '
                "Path or an Intrinsic Function",
            ],
        ),
        (
            {"Error": "error", "ErrorPath": "$.x"},
            ['State Machine.States.x may have only one of ["Error", "ErrorPath"]'],
        ),
        (
            {"Cause": "error", "CausePath": "$.x"},
            ['State Machine.States.x may have only one of ["Cause", "CausePath"]'],
        ),
    ],
)
def test_ng(props, errors):
    state_machine = {
        "States": {"x": {"Type": "Fail", **props}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == errors



from statelint.linter import Linter


def test_ok():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Catch": [
                    {
                        "ErrorEquals": ["States.ALL"],
                        "ResultPath": None,
                        "Next": "y",
                    }
                ],
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_empty_error_equals():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Catch": [{"ErrorEquals": [], "Next": "x"}],
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Catch[0].ErrorEquals is empty, non-empty required"
    ]


def test_empty():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Catch": [{}],
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.Catch[0] does not have required field "ErrorEquals"',
        'State Machine.States.x.Catch[0] does not have required field "Next"',
    ]


def test_states_all_not_by_itself():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Catch": [
                    {
                        "ErrorEquals": ["ValueError", "States.ALL", "KeyError"],
                        "ResultPath": None,
                        "Next": "y",
                    }
                ],
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Catch[0]: States.ALL can only "
        "appear in the last element, and by itself."
    ]


def test_states_all_not_last():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Catch": [
                    {
                        "ErrorEquals": ["States.ALL"],
                        "ResultPath": None,
                        "Next": "y",
                    },
                    {
                        "ErrorEquals": ["ValueError", "KeyError"],
                        "ResultPath": None,
                        "Next": "y",
                    },
                ],
            },
            "y": {"Type": "Succeed"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Catch[0]: States.ALL can only "
        "appear in the last element, and by itself."
    ]



from typing import Any, Optional

from ...fields import TYPE, StateType
from ..choice_state import ChoiceState
from ..fail_state import FailState
from ..map_state import MapState
from ..node import Node, StatePath
from ..parallel_state import ParallelState
from ..pass_state import PassState
from ..succeed_state import SucceedState
from ..task_state import TaskState
from ..unknown_state import UnknownState
from ..wait_state import WaitState
from . import NodeFactory


class StateFactory(NodeFactory):
    def get(self, parent: Node, state_path: StatePath, state: Any) -> Optional[Node]:
        if not isinstance(state, dict):
            return None
        state_type = state.get(TYPE.name)
        if state_type == StateType.SUCCEED.value:
            return SucceedState(state_path, state, parent)
        if state_type == StateType.FAIL.value:
            return FailState(state_path, state, parent)
        if state_type == StateType.PASS.value:
            return PassState(state_path, state, parent)
        if state_type == StateType.WAIT.value:
            return WaitState(state_path, state, parent)
        if state_type == StateType.TASK.value:
            return TaskState(state_path, state, parent)
        if state_type == StateType.CHOICE.value:
            return ChoiceState(state_path, state, parent)
        if state_type == StateType.PARALLEL.value:
            return ParallelState(self, state_path, state, parent)
        if state_type == StateType.MAP.value:
            return MapState(self, state_path, state, parent)

        return UnknownState(state_path, state, parent)



from typing import Optional

from .problem import ProblemType


class Config:
    ignored_problem_types: set[ProblemType]
    is_yaml: bool
    evaluate_jsonata: bool

    @property
    def input_type(self) -> str:
        return "YAML" if self.is_yaml else "JSON"

    def __init__(
        self,
        ignored_problem_types: Optional[set[ProblemType]] = None,
        yaml: bool = False,
        evaluate_jsonata: bool = False,
    ):
        self.ignored_problem_types = ignored_problem_types or set()
        self.is_yaml = yaml
        self.evaluate_jsonata = evaluate_jsonata



import functools
from typing import Pattern

import regex as re  # for unicode codepoint (\p{})

INITIAL_NAME_CLASSES = ["Lu", "Ll", "Lt", "Lm", "Lo", "Nl"]
NON_INITIAL_NAME_CLASSES = ["Mn", "Mc", "Nd", "Pc"]
FOLLOWING_NAME_CLASSES = [*INITIAL_NAME_CLASSES, *NON_INITIAL_NAME_CLASSES]
DOT_SEPARATOR = r"\.\.?"

INTRINSIC_INVOCATION_PATTERN = re.compile(
    r"^States\.(JsonToString|Format|StringToJson|Array|ArrayPartition|ArrayContains"
    r"|ArrayRange|ArrayGetItem|ArrayLength|ArrayUnique|Base64Encode|Base64Decode"
    r"|Hash|JsonMerge|MathRandom|MathAdd|StringSplit|UUID)\(.*\)$"
)


def is_reference_path(value: str) -> bool:
    return bool(_compile_reference_path_pattern().match(value))


def is_path(value: str, is_context_object_ok: bool = False) -> bool:
    if is_context_object_ok:
        value = re.sub(r"^\$\$", "$", value)
    return bool(_compile_path_pattern().match(value))


def is_intrinsic_invocation(value: str) -> bool:
    return isinstance(value, str) and bool(INTRINSIC_INVOCATION_PATTERN.match(value))


@functools.lru_cache()  # TODO
def _compile_reference_path_pattern() -> Pattern[str]:
    name_re = _make_name_re()
    rp_dot_step = DOT_SEPARATOR + name_re
    bracket_step = rf"\['{name_re}']"
    rp_num_index = r"\[\d+]"
    rp_step = f"(({rp_dot_step})|({bracket_step}))({rp_num_index})?"
    reference_path = rf"^\$({rp_step})*$"

    return re.compile(reference_path)


@functools.lru_cache()  # TODO
def _compile_path_pattern() -> Pattern[str]:
    name_re = _make_name_re()
    bracket_step = _make_bracket_step(name_re)

    dot_step = rf"{DOT_SEPARATOR}(({name_re})|(\*))"
    num_index = r"\[\d+(, *\d+)?]"
    star_index = r"\[\*]"
    colon_index = r"\[(-?\d+)?:(-?\d+)?]"
    index = rf"(({num_index})|({star_index})|({colon_index}))"
    step = f"(({dot_step})|({bracket_step})|({index}))({index})?"
    path = rf"^\$({step})*$"
    return re.compile(path)


def _make_name_re() -> str:
    return (
        _classes_to_re(INITIAL_NAME_CLASSES)
        + _classes_to_re(FOLLOWING_NAME_CLASSES)
        + "*"
    )


def _make_bracket_step(name_re: str) -> str:
    return rf"\['{name_re}']"


def _classes_to_re(classes: list[str]) -> str:
    joined = "".join(rf"\p{{{c}}}" for c in classes)
    return f"[{joined}]"



from __future__ import annotations

import functools
from collections import OrderedDict
from typing import Any, NamedTuple, Optional

from ..config import Config
from ..fields import (
    COMMENT,
    QUERY_LANGUAGE,
    Field,
    FieldValue,
    ProblemPredicate,
    QueryLanguage,
)
from ..problem import Problem

StatePathType = str | int | Field


class NameAndPath(NamedTuple):
    name: str
    path: StatePath


class StatePath:
    _root: str
    _paths: list[StatePathType]

    def __init__(self, root: str, *paths: StatePathType):
        self._root = root
        self._paths = list(paths)

    def __str__(self) -> str:
        return functools.reduce(
            lambda x, y: x + self._path_to_str(y), self._paths, self._root
        )

    @staticmethod
    def _path_to_str(path: StatePathType) -> str:
        if isinstance(path, int):
            return f"[{path}]"
        if isinstance(path, Field):
            path = path.name
        return f".{path}"

    @property
    def parent(self) -> StatePath:
        return StatePath(self._root, *self._paths[:-1])

    def make_child(self, *paths: StatePathType) -> StatePath:
        return StatePath(self._root, *self._paths, *paths)

    def make_problem(self, predicate: ProblemPredicate) -> Problem:
        return Problem(f"{self}{predicate}", predicate.type)


def _get_query_language(state: dict[str, Any]) -> Optional[QueryLanguage]:
    raw_query_language = state.get(QUERY_LANGUAGE.name)
    if raw_query_language:
        return QueryLanguage.of(raw_query_language)
    return None


class Node:
    state_path: StatePath
    _state: dict[str, Any]
    _config: Config

    def __init__(
        self,
        state_path: StatePath,
        state: dict[str, Any],
        parent: Node | None,
        config: Config | None = None,
    ) -> None:
        self.state_path = state_path
        self._state = state
        self._query_language = _get_query_language(state) or (
            parent.query_language if parent else QueryLanguage.JSONPath
        )
        self._variable_scopes = parent.variable_scopes if parent else []
        # TODO: This is a bit hacky implementation; needs a structural refactoring.
        maybe_config = parent._config if parent else config
        assert maybe_config
        self._config = maybe_config

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.state_path}, {self._state})"

    @property
    def required_fields(self) -> list[Field]:
        return []

    @property
    def optional_fields(self) -> list[Field]:
        return [COMMENT]

    @property
    def query_language(self) -> QueryLanguage:
        return self._query_language

    @property
    def variable_scopes(self) -> list[dict[str, Any]]:
        return self._variable_scopes

    @property
    def variables(self) -> dict[str, Any]:
        return {k: v for d in self.variable_scopes for k, v in d.items()}

    @property
    def forbidden_fields(self) -> list[Field]:
        # If we just want to validate extra fields, it's enough to put
        # allowed fields to required_fields/optional_fields.
        # Because we want to make our output error messages to be as same to
        # original statelint's as possible, we need this property.
        return []

    def get_reachable_states(self) -> list[NameAndPath]:
        return []

    def get_children(self) -> list[NameAndPath]:
        return []

    def validate(self) -> list[Problem]:
        allowed_fields = OrderedDict(
            (f2.name, f2)
            for f1 in (self.required_fields + self.optional_fields)
            for f2 in f1.get_fields()
        )
        return [
            *self._validate_fields(allowed_fields),
            *self._validate_forbidden_fields(),
            *self._validate_not_allowed_fields(allowed_fields),
            *self._validate_optional_fields(allowed_fields),
            *self._validate_required_fields(),
        ]

    def _validate_fields(self, fields: dict[str, Field]) -> list[Problem]:
        return [
            self.state_path.make_child(field).make_problem(problem)
            for field, value in self._state.items()
            if field in fields
            for problem in fields[field].validate(
                FieldValue(
                    raw_value=value,
                    variables=self.variables,
                    query_language=self.query_language,
                    config=self._config,
                )
            )
        ]

    def _validate_forbidden_fields(self) -> list[Problem]:
        return [
            Problem(f'{self.state_path} has forbidden field "{field}"')
            for field in self.forbidden_fields
            if field.name in self._state
        ]

    def _validate_not_allowed_fields(
        self, allowed_fields: dict[str, Field]
    ) -> list[Problem]:
        forbidden_fields = set(f.name for f in self.forbidden_fields)
        return [
            Problem(f'Field "{field}" not allowed in {self.state_path}')
            for field in self._state
            if field not in allowed_fields and field not in forbidden_fields
        ]

    def _validate_optional_fields(
        self, allowed_fields: dict[str, Field]
    ) -> list[Problem]:
        all_fields = set(self._state)
        return [
            self.state_path.make_problem(problem)
            for field in allowed_fields.values()
            for problem in field.validate_as_optional(all_fields)
        ]

    def _validate_required_fields(self) -> list[Problem]:
        all_fields = set(self._state)
        return [
            self.state_path.make_problem(problem)
            for field in self.required_fields
            for problem in field.validate_as_required(all_fields)
        ]



from ..problem import ProblemPredicate, ProblemType
from .base import Field, NonNullMixin
from .field_value import FieldValue


class ObjectField(NonNullMixin, Field):
    def validate(self, value: FieldValue) -> list[ProblemPredicate]:
        return self.check_type(value, dict, ProblemType.OBJECT)



import pytest

from statelint.linter import Linter


@pytest.mark.parametrize(
    "props",
    [
        {"IntervalSeconds": 1, "MaxAttempts": 0, "JitterStrategy": "FULL"},
        {
            "IntervalSeconds": 99999999,
            "MaxAttempts": 99999998,
            "JitterStrategy": "NONE",
        },
    ],
)
def test_ok(props):
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Retry": [
                    {
                        "ErrorEquals": ["E"],
                        "BackoffRate": 1.0,
                        "MaxDelaySeconds": 4,
                        **props,
                    }
                ],
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_ok_min():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Retry": [{"ErrorEquals": ["E"]}],
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


@pytest.mark.parametrize(
    "props,errors",
    [
        (
            {"ErrorEquals": []},
            [
                "State Machine.States.x.Retry[0].ErrorEquals is empty, non-empty "
                "required",
            ],
        ),
        (
            {"IntervalSeconds": 0},
            [
                "State Machine.States.x.Retry[0].IntervalSeconds is 0 but allowed "
                "floor is 0",
            ],
        ),
        (
            {"MaxAttempts": -1},
            [
                "State Machine.States.x.Retry[0].MaxAttempts is -1 but allowed floor "
                "is -1",
            ],
        ),
        (
            {"MaxAttempts": 99999999},
            [
                "State Machine.States.x.Retry[0].MaxAttempts is 99999999 but allowed "
                "ceiling is 99999999",
            ],
        ),
        (
            {"BackoffRate": 1},
            [
                "State Machine.States.x.Retry[0].BackoffRate is 1 but should be a "
                "Float",
            ],
        ),
        (
            {"MaxDelaySeconds": 0},
            [
                "State Machine.States.x.Retry[0].MaxDelaySeconds is 0 but allowed "
                "floor is 0",
            ],
        ),
        (
            {"JitterStrategy": None},
            [
                "State Machine.States.x.Retry[0].JitterStrategy should be non-null",
                'State Machine.States.x.Retry[0].JitterStrategy is "", not one of the '
                'allowed values ["FULL", "NONE"]',
            ],
        ),
        (
            {"JitterStrategy": "SAMPLE"},
            [
                'State Machine.States.x.Retry[0].JitterStrategy is "SAMPLE", not one '
                'of the allowed values ["FULL", "NONE"]',
            ],
        ),
    ],
)
def test_bad_fields(props, errors):
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Retry": [{"ErrorEquals": ["E"], **props}],
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == errors



from .problem_type import ProblemType


class Problem:
    description: str
    type: ProblemType

    def __init__(self, description: str, _type: ProblemType = ProblemType.UNKNOWN):
        self.description = description
        self.type = _type

    def __str__(self) -> str:
        return self.description

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.description}, {self.type})"



from ..fields import (
    SECONDS,
    SECONDS_PATH,
    TIMESTAMP,
    TIMESTAMP_PATH,
    Field,
    OneOfField,
    QueryLanguage,
)
from .mixins import AssignMixin, NextXorEndMixin, OutputMixin
from .state import State


class WaitState(NextXorEndMixin, OutputMixin, AssignMixin, State):
    @property
    def required_fields(self) -> list[Field]:
        fields = super().required_fields
        if self.query_language == QueryLanguage.JSONata:
            return [*fields, OneOfField(SECONDS, TIMESTAMP)]
        return [
            *fields,
            OneOfField(SECONDS, SECONDS_PATH, TIMESTAMP, TIMESTAMP_PATH),
        ]

    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields]



import pytest

from statelint.config import Config
from statelint.fields.common import QueryLanguage
from statelint.fields.field_value import FieldValue


@pytest.mark.parametrize(
    "query_language,expected",
    [(QueryLanguage.JSONPath, "{% $x %}"), (QueryLanguage.JSONata, 42)],
)
def test_value_with_query_language(query_language, expected):
    field_value = FieldValue(
        raw_value="{% $x %}",
        variables={"x": 42},
        query_language=query_language,
        config=Config(evaluate_jsonata=True),
    )
    assert field_value.value == expected



import pytest

from statelint.config import Config
from statelint.linter import Linter

from .common import get_path


def test_invalid_json_str():
    assert Linter.validate("a") == [
        "Problem reading/parsing JSON: Expecting value: line 1 column 1 (char 0)"
    ]


def test_invalid_json_file():
    file_path = get_path("invalid.json")
    assert Linter.validate(file_path) == [
        "Problem reading/parsing JSON: Expecting ',' delimiter: line 2 "
        "column 11 (char 12)"
    ]


def test_invalid_jsonata_1_file():
    file_path = get_path("invalid-jsonata-1.json")
    assert Linter.validate(file_path) == [
        'Field "Output" not allowed in State Machine.States.FailState'
    ]


@pytest.mark.parametrize(
    "filename",
    [
        ("invalid-jsonata-2.json"),
        ("invalid-jsonata-3.json"),
    ],
)
def test_invalid_jsonata_2_3_file(filename):
    file_path = get_path(filename)
    assert Linter.validate(file_path) == [
        'Field "Assign" not allowed in State Machine.States.JSONPath state',
        'Field "Output" not allowed in State Machine.States.JSONPath state',
        'State Machine.States.JSONata state.Assign is "XXX" but should be an Object',
    ]


def test_null():
    assert Linter.validate("null") == []


@pytest.mark.parametrize(
    "filepath",
    [
        ("states-uuid-invocation.json"),
        ("jsonata-1.json"),
        ("jsonata-2.json"),
        ("jsonata-3.json"),
        ("jsonata-4.json"),
        ("jsonata-5.json"),
    ],
)
def test_valid_json(filepath):
    assert Linter().validate(get_path(filepath)) == []


def test_issue_125_no_eval():
    assert Linter().validate(get_path("invalid-issue-127.json")) == [
        'State Machine.States.MapState.MaxConcurrency is "{% '
        '$number($maxConcurrency) %}" but should be numeric',
    ]


def test_issue_125_eval():
    assert (
        Linter().validate(
            get_path("invalid-issue-127.json"), Config(evaluate_jsonata=True)
        )
        == []
    )



from ..problem import ProblemPredicate
from .any_field import AnyField as _AnyField
from .base import Field
from .bool_field import BoolField as _BoolField
from .common import QueryLanguage, StateType
from .container import OneOfField
from .field_value import FieldValue
from .jsonata_field import JSONataField as _JSONataField
from .list_field import ListField as _ListField
from .list_field import NonEmptyListField as _NonEmptyListField
from .num_field import FloatField as _FloatField
from .num_field import IntegerField as _IntegerField
from .num_field import NumericField as _NumericField
from .object_field import ObjectField as _ObjectField
from .pattern_field import JsonPathField as _JsonPathField
from .pattern_field import NullableRefPathField as _NullableRefPathField
from .pattern_field import RefPathField as _RefPathField
from .pattern_field import RefPathOrFuncField as _RefPathOrFuncField
from .pattern_field import UriField as _UriField
from .str_field import EnumStrField as _EnumStrField
from .str_field import NullableStrField as _NullableStrField
from .str_field import StrField as _StrField
from .timestamp_field import TimestampField as _TimestampField

VERSION = _StrField("Version")
COMMENT = _StrField("Comment")
QUERY_LANGUAGE = _EnumStrField("QueryLanguage", ["JSONPath", "JSONata"])
OUTPUT = _AnyField("Output")
ASSIGN = _ObjectField("Assign")

SECONDS = _IntegerField("Seconds", 0, inclusive=False)
SECONDS_PATH = _RefPathField("SecondsPath")
TIMESTAMP = _TimestampField("Timestamp")
TIMESTAMP_PATH = _RefPathField("TimestampPath")

CAUSE = _StrField("Cause")
CAUSE_PATH = _RefPathOrFuncField("CausePath")
ERROR = _StrField("Error")
ERROR_PATH = _RefPathOrFuncField("ErrorPath")

RESULT = _AnyField("Result")
RESULT_SELECTOR = _AnyField("ResultSelector")
RESULT_PATH = _NullableRefPathField("ResultPath")
PARAMETERS = _AnyField("Parameters")

NEXT = _StrField("Next")
END = _BoolField("End")

STATES = _ObjectField("States")
START_AT = _StrField("StartAt")

TIMEOUT_SECONDS = _IntegerField("TimeoutSeconds", 0, 99999999)
HEARTBEAT_SECONDS = _IntegerField("HeartbeatSeconds", 0, 99999999)
TIMEOUT_SECONDS_PATH = _RefPathField("TimeoutSecondsPath")
HEARTBEAT_SECONDS_PATH = _RefPathField("HeartbeatSecondsPath")
CREDENTIALS = _ObjectField("Credentials")

TYPE = _EnumStrField("Type", [t.value for t in StateType])

INPUT_PATH = _NullableStrField("InputPath")
OUTPUT_PATH = _NullableStrField("OutputPath")

RESOURCE = _UriField("Resource")
ARGUMENTS = _JSONataField("Arguments", _ObjectField)

CATCH = _ListField("Catch", _ObjectField)
RETRY = _ListField("Retry", _ObjectField)
ERROR_EQUALS = _NonEmptyListField("ErrorEquals", _StrField)
INTERVAL_SECONDS = _IntegerField("IntervalSeconds", 0)
MAX_ATTEMPTS = _IntegerField("MaxAttempts", -1, 99999999)
BACKOFF_RATE = _FloatField("BackoffRate", 1, inclusive=False)
MAX_DELAY_SECONDS = _IntegerField("MaxDelaySeconds", 0)
JITTER_STRATEGY = _EnumStrField("JitterStrategy", ["FULL", "NONE"])


CHOICES = _NonEmptyListField("Choices", _ObjectField)
DEFAULT = _StrField("Default")
CONDITION = _StrField("Condition")

AND = _NonEmptyListField("And", _ObjectField)
OR = _NonEmptyListField("Or", _ObjectField)
NOT = _ObjectField("Not")

VARIABLE = _AnyField("Variable")

STRING_EQUALS = _StrField("StringEquals")
STRING_LESS_THAN = _StrField("StringLessThan")
STRING_GREATER_THAN = _StrField("StringGreaterThan")
STRING_LESS_THAN_EQUALS = _StrField("StringLessThanEquals")
STRING_GREATER_THAN_EQUALS = _StrField("StringGreaterThanEquals")
NUMERIC_EQUALS = _NumericField("NumericEquals")
NUMERIC_LESS_THAN = _NumericField("NumericLessThan")
NUMERIC_GREATER_THAN = _NumericField("NumericGreaterThan")
NUMERIC_LESS_THAN_EQUALS = _NumericField("NumericLessThanEquals")
NUMERIC_GREATER_THAN_EQUALS = _NumericField("NumericGreaterThanEquals")
BOOLEAN_EQUALS = _BoolField("BooleanEquals")
TIMESTAMP_EQUALS = _TimestampField("TimestampEquals")
TIMESTAMP_LESS_THAN = _TimestampField("TimestampLessThan")
TIMESTAMP_GREATER_THAN = _TimestampField("TimestampGreaterThan")
TIMESTAMP_LESS_THAN_EQUALS = _TimestampField("TimestampLessThanEquals")
TIMESTAMP_GREATER_THAN_EQUALS = _TimestampField("TimestampGreaterThanEquals")
STRING_EQUALS_PATH = _RefPathField("StringEqualsPath")
STRING_LESS_THAN_PATH = _RefPathField("StringLessThanPath")
STRING_GREATER_THAN_PATH = _RefPathField("StringGreaterThanPath")
STRING_LESS_THAN_EQUALS_PATH = _RefPathField("StringLessThanEqualsPath")
STRING_GREATER_THAN_EQUALS_PATH = _RefPathField("StringGreaterThanEqualsPath")
NUMERIC_EQUALS_PATH = _RefPathField("NumericEqualsPath")
NUMERIC_LESS_THAN_PATH = _RefPathField("NumericLessThanPath")
NUMERIC_GREATER_THAN_PATH = _RefPathField("NumericGreaterThanPath")
NUMERIC_LESS_THAN_EQUALS_PATH = _RefPathField("NumericLessThanEqualsPath")
NUMERIC_GREATER_THAN_EQUALS_PATH = _RefPathField("NumericGreaterThanEqualsPath")
BOOLEAN_EQUALS_PATH = _RefPathField("BooleanEqualsPath")
TIMESTAMP_EQUALS_PATH = _RefPathField("TimestampEqualsPath")
TIMESTAMP_LESS_THAN_PATH = _RefPathField("TimestampLessThanPath")
TIMESTAMP_GREATER_THAN_PATH = _RefPathField("TimestampGreaterThanPath")
TIMESTAMP_LESS_THAN_EQUALS_PATH = _RefPathField("TimestampLessThanEqualsPath")
TIMESTAMP_GREATER_THAN_EQUALS_PATH = _RefPathField("TimestampGreaterThanEqualsPath")
IS_NULL = _BoolField("IsNull")
IS_PRESENT = _BoolField("IsPresent")
IS_NUMERIC = _BoolField("IsNumeric")
IS_STRING = _BoolField("IsString")
IS_BOOLEAN = _BoolField("IsBoolean")
IS_TIMESTAMP = _BoolField("IsTimestamp")
STRING_MATCHES = _StrField("StringMatches")

BRANCHES = _ListField("Branches", _ObjectField)

ITERATOR = _ObjectField("Iterator")
ITEMS_PATH = _JsonPathField("ItemsPath")
MAX_CONCURRENCY = _NumericField("MaxConcurrency", 0, inclusive=False)
MAX_CONCURRENCY_PATH = _RefPathField("MaxConcurrencyPath")

ITEM_PROCESSOR = _ObjectField("ItemProcessor")
PROCESSOR_CONFIG = _ObjectField("ProcessorConfig")
ITEM_SELECTOR = _AnyField("ItemSelector")
TOLERATED_FAILURE_COUNT = _IntegerField("ToleratedFailureCount", 0, inclusive=False)
TOLERATED_FAILURE_COUNT_PATH = _RefPathField("ToleratedFailureCountPath")
ITEM_READER = _ObjectField("ItemReader")
READER_CONFIG = _ObjectField("ReaderConfig")
MAX_ITEMS = _IntegerField("MaxItems", 0)
MAX_ITEMS_PATH = _RefPathField("MaxItemsPath")
RESULT_WRITER = _ObjectField("ResultWriter")
ITEM_BATCHER = _ObjectField("ItemBatcher")
BATCH_INPUT = _AnyField("BatchInput")
MAX_ITEMS_PER_BATCH = _IntegerField("MaxItemsPerBatch", 0)
MAX_ITEMS_PER_BATCH_PATH = _RefPathField("MaxItemsPerBatchPath")
MAX_INPUT_BYTES_PER_BATCH = _IntegerField("MaxInputBytesPerBatch", 0)
MAX_INPUT_BYTES_PER_BATCH_PATH = _RefPathField("MaxInputBytesPerBatchPath")
TOLERATED_FAILURE_PERCENTAGE = _IntegerField(
    "ToleratedFailurePercentage", 0, 100, inclusive=False
)
TOLERATED_FAILURE_PERCENTAGE_PATH = _RefPathField("ToleratedFailurePercentagePath")
ITEMS = _JSONataField("Items", lambda s: _ListField(s, _AnyField))



from collections import OrderedDict
from typing import Any, Optional

from ..config import Config
from ..fields import ASSIGN, END, NEXT, START_AT, STATES, TYPE, Field, StateType
from ..problem import Problem
from .factory import NodeFactory
from .node import NameAndPath, Node, StatePath


class ContainerState(Node):
    def __init__(
        self,
        node_factory: NodeFactory,
        state_path: StatePath,
        state: dict[str, Any],
        parent: Node | None,
        config: Config | None = None,
    ) -> None:
        super().__init__(state_path, state, parent, config)
        self._node_factory = node_factory
        self._variable_scopes = [
            *(parent.variable_scopes if parent else []),
            self._collect_variables(state),
        ]
        self._states = self._collect_states()

    def _collect_states(self) -> dict[str, Node]:
        ret_dict = OrderedDict()
        states = self._state.get(STATES.name)
        if isinstance(states, dict):
            for state_name, state in states.items():
                validator = self._node_factory.get(
                    self,
                    self.state_path.make_child(STATES, state_name),
                    state,
                )
                if validator:
                    ret_dict[state_name] = validator
        return ret_dict

    @staticmethod
    def _collect_variables(this: dict[str, Any]) -> dict[str, Any]:
        states = this.get(STATES.name)
        if not isinstance(states, dict):
            return {}

        return dict(
            variables
            for node in states.values()
            if isinstance(node, dict)
            for assign in [node.get(ASSIGN.name)]
            if isinstance(assign, dict)
            for variables in assign.items()
        )

    @property
    def required_fields(self) -> list[Field]:
        return [*super().required_fields, STATES, START_AT]

    def get_children(self) -> list[NameAndPath]:
        return [NameAndPath(n, s.state_path) for n, s in self._states.items()]

    def validate(self) -> list[Problem]:
        problems = super().validate()
        states = self._state.get(STATES.name)
        reachable_states = set()
        if isinstance(states, dict):
            all_states: dict[str, StatePath] = {
                k: s.state_path for k, s in self._states.items()
            }
            for state in self._states.values():
                problems.extend(state.validate())
                _reachable_states = state.get_reachable_states()
                for s in (s for s in _reachable_states if s.name not in states):
                    problems.append(
                        Problem(
                            f'No state found named "{s.name}", referenced at {s.path}'
                        )
                    )
                reachable_states.update(_reachable_states)
                for child in state.get_children():
                    if child.name in all_states:
                        problems.append(
                            Problem(
                                f'State "{child.name}", defined at '
                                f"{child.path.parent}, is also defined at "
                                f"{all_states[child.name].parent}"
                            )
                        )
                    else:
                        all_states[child.name] = state.state_path

        problems += self._validate_transition(
            self._state.get(START_AT.name), states, reachable_states
        )
        return problems

    def _validate_transition(
        self, start_at: Optional[str], states: Any, reachable_states: set[NameAndPath]
    ) -> list[Problem]:
        if not isinstance(states, dict):
            return []
        if start_at not in states:
            problems = [
                Problem(f"No transition found to state {self.state_path.make_child(s)}")
                for s in states
            ]
            if start_at:
                problems = [
                    Problem(
                        f"{START_AT.name} value {start_at} not found in "
                        f"{STATES.name} field at {self.state_path}"
                    )
                ] + problems
            return problems
        state_names = list(states)

        problems = []
        terminal_found = False
        for state_name in (start_at, *(s.name for s in reachable_states)):
            while True:
                if state_name not in state_names:
                    break  # checked already.
                state_names.remove(state_name)
                state = states[state_name]
                if not isinstance(state, dict):
                    break
                if _is_terminal(state):
                    terminal_found = True
                if NEXT.name not in state:
                    break
                next_state_name = state[NEXT.name]
                if next_state_name not in states:
                    assert state_name, "state_names contains None"
                    state_path = self.state_path.make_child(STATES, state_name, NEXT)
                    problems.append(
                        Problem(
                            f'No state found named "{next_state_name}", '
                            f"referenced at {state_path}"
                        )
                    )
                    break
                state_name = next_state_name

        if state_names:
            problems.extend(
                Problem(f"No transition found to state {self.state_path.make_child(s)}")
                for s in state_names
            )

        if not terminal_found:
            problems.append(
                Problem(
                    "No terminal state found in machine at "
                    f"{self.state_path.make_child(STATES)}"
                )
            )

        return problems


def _is_terminal(state: dict[str, Any]) -> bool:
    if state.get(TYPE.name) in (StateType.SUCCEED.value, StateType.FAIL.value):
        return True
    if state.get(END.name):
        return True

    return False



from typing import Any

from ..config import Config
from ..fields import QUERY_LANGUAGE, VERSION, Field
from .container_state import ContainerState
from .factory import NodeFactory
from .mixins import TimeoutSecondsMixin
from .node import StatePath


class StateMachine(TimeoutSecondsMixin, ContainerState):
    def __init__(
        self, node_factory: NodeFactory, state: dict[str, Any], config: Config
    ) -> None:
        super().__init__(node_factory, StatePath("State Machine"), state, None, config)

    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields, VERSION, QUERY_LANGUAGE]



from ...fields import ASSIGN, Field, QueryLanguage
from ..node import Node


class AssignMixin(Node):
    @property
    def optional_fields(self) -> list[Field]:
        fields = super().optional_fields
        if self.query_language == QueryLanguage.JSONata:
            return [*fields, ASSIGN]
        return fields



from ..fields import INPUT_PATH, OUTPUT_PATH, QUERY_LANGUAGE, TYPE, Field, QueryLanguage
from .node import Node


class State(Node):
    @property
    def required_fields(self) -> list[Field]:
        return [*super().required_fields, TYPE]

    @property
    def optional_fields(self) -> list[Field]:
        fields = [*super().optional_fields, QUERY_LANGUAGE]
        if self.query_language == QueryLanguage.JSONata:
            return fields
        return [*fields, INPUT_PATH, OUTPUT_PATH]



from dateutil import parser

from ..problem import ProblemPredicate
from .base import Field, NonNullMixin
from .common import to_json
from .field_value import FieldValue


class TimestampField(NonNullMixin, Field):
    def validate(self, field_value: FieldValue) -> list[ProblemPredicate]:
        problems = super().validate(field_value)
        if problems:
            return problems
        value = field_value.value
        if isinstance(value, str) and _is_rfc3339_format(value):
            return []

        return [
            ProblemPredicate(f" is {to_json(value)} but should be an RFC3339 timestamp")
        ]


def _is_rfc3339_format(value: str) -> bool:
    # ISO8601/RFC3339
    try:
        parser.isoparse(value)
    except ValueError:
        return False
    return True



import itertools
import json
import os.path
from unittest import mock
from unittest.mock import call

import pytest

from statelint.cli import main


@pytest.mark.parametrize(
    "arg_0,expected_exit_code,problems",
    [
        ({"StartAt": "x", "States": {"x": {"Type": "Fail"}}}, 0, []),
        (
            {"StartAt": "x", "States": {}},
            1,
            [
                "One error:",
                " StartAt value x not found in States field at State Machine",
            ],
        ),
        (
            {"StartAt": "x", "States": {"x": {"Type": "Task"}}},
            1,
            [
                "2 errors:",
                ' State Machine.States.x does not have required field "Next"',
                " No terminal state found in machine at State Machine.States",
            ],
        ),
    ],
)
@mock.patch("builtins.print")
def test_main(mocked_print, arg_0, expected_exit_code, problems):
    exit_code = main([json.dumps(arg_0)])
    assert mocked_print.mock_calls == [call(p) for p in problems]
    assert exit_code == expected_exit_code


@pytest.mark.parametrize(
    "arg_1,problems",
    [
        (
            "--ignore=FLOAT",
            [
                "One error:",
                ' State Machine.States.x.Resource is "x" but should be A URI',
            ],
        ),
        (
            "--ignore=URI",
            [
                "One error:",
                " State Machine.States.x.Retry[0].BackoffRate is 1 but should be "
                "a Float",
            ],
        ),
        ("--ignore=FLOAT,URI", []),
        ("--ignore=URI,FLOAT", []),
    ],
)
@mock.patch("builtins.print")
def test_ignore(mocked_print, arg_1, problems):
    arg_0 = json.dumps(
        {
            "StartAt": "x",
            "States": {
                "x": {
                    "Type": "Task",
                    "Resource": "x",
                    "End": True,
                    "Retry": [{"ErrorEquals": ["e"], "BackoffRate": 1}],
                }
            },
        }
    )
    main([arg_0, arg_1])
    assert mocked_print.mock_calls == [call(p) for p in problems]


@pytest.mark.parametrize("ignore", ["", "ARRAY"])
def test_invalid_ignore(ignore):
    with pytest.raises(SystemExit):
        main(["{}", f"--ignore={ignore}"])


def test_yaml():
    yaml_path = os.path.join(
        os.path.dirname(__file__), "data", "batch-job-notification.yaml"
    )
    assert main([yaml_path, "--yaml"]) == 0


@mock.patch("builtins.print")
@mock.patch("statelint.cli.importlib.import_module")
def test_yaml_not_installed(import_module, *_):
    import_module.side_effect = ImportError()
    assert main(["{}", "--yaml"]) == 2


@mock.patch("statelint.cli.sys.argv", ["", "{}"])
@mock.patch("builtins.print")
def test_default_argv(mocked_print):
    assert main() == 1
    assert mocked_print.mock_calls == [
        call("2 errors:"),
        call(' State Machine does not have required field "States"'),
        call(' State Machine does not have required field "StartAt"'),
    ]


@mock.patch("builtins.print")
@mock.patch("statelint.cli.importlib.import_module")
def test_jsonata_not_installed(import_module, *_):
    import_module.side_effect = ImportError()
    assert main(["{}", "--eval-jsonata"]) == 2


@mock.patch("builtins.print")
def test_multiple_input(mocked_print):
    args = ['{"Comment":"1"}', "{}"]
    assert main(args) == 1

    assert mocked_print.mock_calls == list(
        itertools.chain.from_iterable(
            [
                call(arg),
                call("2 errors:"),
                call(' State Machine does not have required field "States"'),
                call(' State Machine does not have required field "StartAt"'),
            ]
            for arg in args
        )
    )


@pytest.mark.parametrize(
    "flags,exit_code,print_calls",
    [
        (["--eval-jsonata"], 0, []),
        (
            [],
            1,
            [
                call("One error:"),
                call(
                    " State Machine.States.MapState.MaxConcurrency is "
                    '"{% $number($maxConcurrency) %}" but should be numeric'
                ),
            ],
        ),
    ],
)
@mock.patch("builtins.print")
def test_eval_jsonata(mocked_print, flags, exit_code, print_calls):
    file_path = os.path.join(
        os.path.dirname(__file__), "data", "invalid-issue-127.json"
    )
    assert main([file_path, *flags]) == exit_code
    assert mocked_print.mock_calls == print_calls



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Any

from ..config import Config
from ..utils.jsonata_helper import evaluate_jsonata, extract_jsonata_exp
from .common import QueryLanguage


@dataclass
class FieldValue:
    raw_value: Any
    query_language: QueryLanguage
    variables: dict[str, Any]
    config: Config

    def of(self, value: Any) -> FieldValue:
        return replace(self, raw_value=value)

    @property
    def value(self) -> Any:
        value = self.raw_value
        if not self.config.evaluate_jsonata:
            return value

        if self.query_language != QueryLanguage.JSONata:
            return value

        if not isinstance(value, str):
            return value

        exp = extract_jsonata_exp(value)
        if not exp:
            return value

        # TODO: Ideally, when correctly evaluated, we should return None.
        # However, since we are too lazy to handle that properly, we currently
        # just return the original value in all cases.
        return evaluate_jsonata(exp, self.variables) or value



import re
from typing import Any

_PATTERN = r"\{% +(.+) +%\}"


def extract_jsonata_exp(value: str) -> str | None:
    match = re.match(_PATTERN, value)
    return match and match[1]


def is_jsonata_value(value: str) -> bool:
    return bool(extract_jsonata_exp(value))


def evaluate_jsonata(expr_str: str, variable: dict[str, Any]) -> Any:
    from jsonata import Jsonata  # type: ignore

    expr = Jsonata(expr_str)
    for k, v in variable.items():
        expr.assign(k, v)

    try:
        return expr.evaluate(variable)
    except Exception:
        return None



import pytest

from statelint.linter import Linter


def test_ok():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "ResultPath": None,
                "Resource": "arn:x",
                "TimeoutSeconds": 1,
                "ResultSelector": {},
                "Parameters": {},
                "Credentials": {},
                "Catch": [],
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_ok_min():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == []


def test_bad_fields():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "ResultPath": ".x",
                "Resource": "xxx",
                "TimeoutSeconds": 0,
                "ResultSelector": None,
                "Parameters": "x",
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        'State Machine.States.x.ResultPath is ".x" but should be a Reference Path',
        'State Machine.States.x.Resource is "xxx" but should be A URI',
        "State Machine.States.x.TimeoutSeconds is 0 but allowed floor is 0",
    ]


@pytest.mark.parametrize("name", ["TimeoutSeconds", "HeartbeatSeconds"])
def test_one_of_fields(name):
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                name: 1,
                f"{name}Path": "$",
            }
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        f'State Machine.States.x may have only one of ["{name}", "{name}Path"]',
    ]


def test_catch():
    state_machine = {
        "States": {"x": {"Type": "Task", "End": True, "Resource": "arn:x", "Catch": 1}},
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Catch is 1 but should be an Array"
    ]


def test_catch_null():
    state_machine = {
        "States": {
            "x": {"Type": "Task", "End": True, "Resource": "arn:x", "Catch": None}
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Catch should be non-null"
    ]


def test_catch_element():
    state_machine = {
        "States": {
            "x": {
                "Type": "Task",
                "End": True,
                "Resource": "arn:x",
                "Catch": [1, "x"],
            },
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "State Machine.States.x.Catch[0] is 1 but should be an Object",
        'State Machine.States.x.Catch[1] is "x" but should be an Object',
    ]



import itertools
from typing import Any

from ...fields import (
    BATCH_INPUT,
    ITEM_SELECTOR,
    PARAMETERS,
    RESULT_SELECTOR,
    Field,
    OneOfField,
    QueryLanguage,
)
from ...problem import Problem
from ...utils.re_helper import is_intrinsic_invocation, is_path
from ..node import Node, StatePath


class ParametersMixin(Node):
    @property
    def optional_fields(self) -> list[Field]:
        fields = super().optional_fields
        if self.query_language == QueryLanguage.JSONata:
            return fields
        return [*fields, PARAMETERS]

    def validate(self) -> list[Problem]:
        problems = [*super().validate()]
        if not isinstance(self._state.get(PARAMETERS.name), dict):
            return problems

        return problems + _validate_payload(
            PARAMETERS, self.state_path, self._state[PARAMETERS.name]
        )


class ItemSelectorMixin(Node):
    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields, OneOfField(PARAMETERS, ITEM_SELECTOR)]

    def validate(self) -> list[Problem]:
        problems = [*super().validate()]
        for field in (PARAMETERS, ITEM_SELECTOR):
            if isinstance(self._state.get(field.name), dict):
                problems += _validate_payload(
                    field, self.state_path, self._state[field.name]
                )

        return problems


class ResultSelectorMixin(Node):
    @property
    def optional_fields(self) -> list[Field]:
        fields = super().optional_fields
        if self.query_language == QueryLanguage.JSONata:
            return fields
        return [*fields, RESULT_SELECTOR]

    def validate(self) -> list[Problem]:
        problems = [*super().validate()]
        if not isinstance(self._state.get(RESULT_SELECTOR.name), dict):
            return problems

        return problems + _validate_payload(
            RESULT_SELECTOR, self.state_path, self._state[RESULT_SELECTOR.name]
        )


class BatchInputMixin(Node):
    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields, BATCH_INPUT]

    def validate(self) -> list[Problem]:
        problems = super().validate()
        batch_input = self._state.get(BATCH_INPUT.name)
        if not isinstance(batch_input, dict):
            return problems

        return problems + _validate_payload(BATCH_INPUT, self.state_path, batch_input)


def _validate_payload(
    field: Field, current_path: StatePath, param: Any
) -> list[Problem]:
    problems = []
    if isinstance(param, list):
        return list(
            itertools.chain.from_iterable(
                _validate_payload(field, current_path.make_child(i), p)
                for i, p in enumerate(param)
            )
        )
    if not isinstance(param, dict):
        return []

    for key, value in param.items():
        if not key.endswith(".$"):
            problems.extend(
                _validate_payload(field, current_path.make_child(key), value)
            )
            continue

        if isinstance(value, str) and _is_valid_parameter(value):
            continue
        problems.append(
            Problem(
                f'Field "{key}" of {field} at "{current_path}" is not '
                "a JSONPath or intrinsic function expression"
            )
        )
    return problems


def _is_valid_parameter(value: str) -> bool:
    return is_path(value, True) or is_intrinsic_invocation(value)



from typing import Any, Optional

from ..fields import (
    ITEM_BATCHER,
    ITEM_PROCESSOR,
    ITEM_READER,
    ITEMS,
    ITEMS_PATH,
    ITERATOR,
    MAX_CONCURRENCY,
    MAX_CONCURRENCY_PATH,
    MAX_INPUT_BYTES_PER_BATCH,
    MAX_INPUT_BYTES_PER_BATCH_PATH,
    MAX_ITEMS,
    MAX_ITEMS_PATH,
    MAX_ITEMS_PER_BATCH,
    MAX_ITEMS_PER_BATCH_PATH,
    PROCESSOR_CONFIG,
    READER_CONFIG,
    RESOURCE,
    RESULT_WRITER,
    TOLERATED_FAILURE_COUNT,
    TOLERATED_FAILURE_COUNT_PATH,
    TOLERATED_FAILURE_PERCENTAGE,
    TOLERATED_FAILURE_PERCENTAGE_PATH,
    Field,
    QueryLanguage,
)
from ..fields.container import OneOfField
from ..problem import Problem
from .container_state import ContainerState
from .factory import NodeFactory
from .mixins import (
    ArgumentsMixin,
    AssignMixin,
    BatchInputMixin,
    CatchMixin,
    ItemSelectorMixin,
    NextXorEndMixin,
    OutputMixin,
    ParametersMixin,
    ResultPathMixin,
    ResultSelectorMixin,
    RetryMixin,
    TimeoutSecondsMixin,
)
from .node import NameAndPath, Node, StatePath
from .state import State


class MapConfig(ParametersMixin, ArgumentsMixin):
    @property
    def required_fields(self) -> list[Field]:
        return [*super().required_fields, RESOURCE]


class ReaderConfig(Node):
    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields, OneOfField(MAX_ITEMS, MAX_ITEMS_PATH)]


class ItemReader(MapConfig):
    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields, READER_CONFIG]

    def validate(self) -> list[Problem]:
        problems = super().validate()
        config = self._state.get(READER_CONFIG.name)
        if not isinstance(config, dict):
            return problems

        return (
            problems
            + ReaderConfig(
                self.state_path.make_child(READER_CONFIG), config, self
            ).validate()
        )


class ResultWriter(MapConfig):
    pass


class ItemProcessor(ContainerState):
    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields, PROCESSOR_CONFIG]


class ItemBatcher(BatchInputMixin):
    @property
    def optional_fields(self) -> list[Field]:
        return [
            *super().optional_fields,
            OneOfField(MAX_ITEMS_PER_BATCH, MAX_ITEMS_PER_BATCH_PATH),
            OneOfField(MAX_INPUT_BYTES_PER_BATCH, MAX_INPUT_BYTES_PER_BATCH_PATH),
        ]


class MapState(
    NextXorEndMixin,
    ResultPathMixin,
    ItemSelectorMixin,
    ResultSelectorMixin,
    TimeoutSecondsMixin,
    CatchMixin,
    RetryMixin,
    OutputMixin,
    AssignMixin,
    State,
):
    def __init__(
        self,
        node_factory: NodeFactory,
        state_path: StatePath,
        state: dict[str, Any],
        parent: Node,
    ) -> None:
        super().__init__(state_path, state, parent)
        self._node_factory = node_factory
        self._iterator = self._get_iterator(state)
        self._validators = (
            x
            for x in (
                self._iterator,
                self._get_item_reader(state),
                self._get_result_writer(state),
                self._get_item_batcher(state),
            )
            if x
        )

    @property
    def required_fields(self) -> list[Field]:
        return [*super().required_fields, OneOfField(ITERATOR, ITEM_PROCESSOR)]

    @property
    def optional_fields(self) -> list[Field]:
        fields = [
            *super().optional_fields,
            ITEM_READER,
            RESULT_WRITER,
            ITEM_BATCHER,
        ]
        if self.query_language == QueryLanguage.JSONata:
            return [
                *fields,
                ITEMS,
                MAX_CONCURRENCY,
                TOLERATED_FAILURE_COUNT,
                TOLERATED_FAILURE_PERCENTAGE,
            ]
        return [
            *fields,
            ITEMS_PATH,
            OneOfField(MAX_CONCURRENCY, MAX_CONCURRENCY_PATH),
            OneOfField(TOLERATED_FAILURE_COUNT, TOLERATED_FAILURE_COUNT_PATH),
            OneOfField(TOLERATED_FAILURE_PERCENTAGE, TOLERATED_FAILURE_PERCENTAGE_PATH),
        ]

    def get_children(self) -> list[NameAndPath]:
        if not self._iterator:
            return []
        return list(self._iterator.get_children())

    def validate(self) -> list[Problem]:
        return super().validate() + [y for x in self._validators for y in x.validate()]

    def _get_iterator(self, state: dict[str, Any]) -> Optional[ContainerState]:
        iterator = state.get(ITERATOR.name)
        item_processor = state.get(ITEM_PROCESSOR.name)
        if isinstance(iterator, dict):
            return ContainerState(
                self._node_factory, self.state_path.make_child(ITERATOR), iterator, self
            )
        if isinstance(item_processor, dict):
            return ItemProcessor(
                self._node_factory,
                self.state_path.make_child(ITEM_PROCESSOR),
                item_processor,
                self,
            )
        return None

    def _get_item_reader(self, state: dict[str, Any]) -> Optional[ItemReader]:
        item_reader = state.get(ITEM_READER.name)
        if isinstance(item_reader, dict):
            return ItemReader(
                self.state_path.make_child(ITEM_READER), item_reader, self
            )
        return None

    def _get_result_writer(self, state: dict[str, Any]) -> Optional[ResultWriter]:
        result_writer = state.get(RESULT_WRITER.name)
        if isinstance(result_writer, dict):
            return ResultWriter(
                self.state_path.make_child(RESULT_WRITER), result_writer, self
            )
        return None

    def _get_item_batcher(self, state: dict[str, Any]) -> Optional[ItemBatcher]:
        item_batcher = state.get(ITEM_BATCHER.name)
        if isinstance(item_batcher, dict):
            return ItemBatcher(
                self.state_path.make_child(ITEM_BATCHER), item_batcher, self
            )
        return None



from .mixins import (
    AssignMixin,
    NextXorEndMixin,
    OutputMixin,
    ParametersMixin,
    ResultMixin,
    ResultPathMixin,
)
from .state import State


class PassState(
    ResultPathMixin,
    ResultMixin,
    ParametersMixin,
    NextXorEndMixin,
    OutputMixin,
    AssignMixin,
    State,
):
    pass



import re
from abc import ABC, abstractmethod
from typing import Callable

from ..problem import ProblemPredicate, ProblemType
from ..utils.re_helper import is_intrinsic_invocation, is_path, is_reference_path
from .base import NonNullMixin
from .field_value import FieldValue
from .str_field import NullableStrField


class PatternField(NullableStrField, ABC):
    def __init__(self, name: str, is_match: Callable[[str], bool]) -> None:
        super().__init__(name)
        self._is_match = is_match

    @property
    @abstractmethod
    def problem_type(self) -> ProblemType:
        pass  # pragma: no cover

    def validate(self, field_value: FieldValue) -> list[ProblemPredicate]:
        problems = super().validate(field_value)
        value = field_value.value
        if problems:
            return problems
        if isinstance(value, str) and not self._is_match(value):
            return [
                ProblemPredicate(
                    f' is "{value}" but should be {self.problem_type}',
                    self.problem_type,
                )
            ]
        return []


class UriField(NonNullMixin, PatternField):
    def __init__(self, name: str) -> None:
        super().__init__(name, lambda x: bool(re.compile(r"^[a-z]+:").match(x)))

    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.URI


class NullableRefPathField(PatternField):
    def __init__(self, name: str) -> None:
        super().__init__(name, is_reference_path)

    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.REFERENCE_PATH


class RefPathField(NonNullMixin, NullableRefPathField):
    pass


class NullableRefPathOrFuncField(PatternField):
    def __init__(self, name: str) -> None:
        super().__init__(
            name, lambda v: is_reference_path(v) or is_intrinsic_invocation(v)
        )

    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.REFERENCE_PATH_OR_FUNC


class RefPathOrFuncField(NonNullMixin, NullableRefPathOrFuncField):
    pass


class JsonPathField(NonNullMixin, PatternField):
    def __init__(self, name: str) -> None:
        super().__init__(name, lambda x: is_path(x, True))

    @property
    def problem_type(self) -> ProblemType:
        return ProblemType.JSON_PATH



from unittest.mock import Mock

import pytest

from statelint.config import Config
from statelint.linter import Linter
from statelint.nodes.state_machine import StateMachine


def test_ok():
    assert Linter.validate({"States": {"x": {"Type": "Succeed"}}, "StartAt": "x"}) == []


def test_required_fields():
    assert Linter.validate({"Comment": "xxx", "Version": "yyy"}) == [
        'State Machine does not have required field "States"',
        'State Machine does not have required field "StartAt"',
    ]


def test_not_allowed_fields():
    statemachine = {"States": {"x": {"Type": "Succeed"}}, "StartAt": "x", "a": 1}
    assert Linter.validate(statemachine) == [
        'Field "a" not allowed in State Machine',
    ]


def test_type_error():
    statemachine = {"States": "1", "StartAt": {"x": 1}}
    assert Linter.validate(statemachine) == [
        'State Machine.States is "1" but should be an Object',
        'State Machine.StartAt is {"x": 1} but should be a String',
    ]


@pytest.mark.parametrize(
    "timeout_sec,expected",
    [
        (0, ["State Machine.TimeoutSeconds is 0 but allowed floor is 0"]),
        (1.2, ["State Machine.TimeoutSeconds is 1.2 but should be an Integer"]),
        (42, []),
        (
            99999999,
            [
                "State Machine.TimeoutSeconds is 99999999 but "
                "allowed ceiling is 99999999"
            ],
        ),
    ],
)
def test_positive_int_fields(timeout_sec, expected):
    statemachine = {
        "States": {"x": {"Type": "Succeed"}},
        "StartAt": "x",
        "TimeoutSeconds": timeout_sec,
    }
    assert Linter.validate(statemachine) == expected


def test_no_start_at():
    state_machine = {"States": {"x": {"Type": "Succeed"}, "y": 1}}
    assert Linter.validate(state_machine) == [
        'State Machine does not have required field "StartAt"',
        "No transition found to state State Machine.x",
        "No transition found to state State Machine.y",
    ]


def test_bad_start_at():
    state_machine = {"States": {"y": {"Type": "Succeed"}}, "StartAt": "x"}
    assert Linter.validate(state_machine) == [
        "StartAt value x not found in States field at State Machine",
        "No transition found to state State Machine.y",
    ]


def test_no_transion():
    state_machine = {"States": {"x": {"Type": "Succeed"}, "y": 1}, "StartAt": "x"}
    assert Linter.validate(state_machine) == [
        "No transition found to state State Machine.y"
    ]


def test_loop():
    state_machine = {
        "States": {
            "x": {"Type": "Pass", "Next": "y"},
            "y": {"Type": "Pass", "Next": "x"},
        },
        "StartAt": "x",
    }
    assert Linter.validate(state_machine) == [
        "No terminal state found in machine at State Machine.States"
    ]


def test_no_terminal_state():
    state_machine = {"States": {"x": 1}, "StartAt": "x"}
    assert Linter.validate(state_machine) == [
        "No terminal state found in machine at State Machine.States"
    ]


def test_null_state():
    state_machine = {"States": {"x": None}, "StartAt": "x"}
    assert Linter.validate(state_machine) == [
        "No terminal state found in machine at State Machine.States"
    ]


def test_repr():
    assert (
        f'{StateMachine(Mock(), {"x": 42}, Config())}'
        == "StateMachine(State Machine, {'x': 42})"
    )



from ...fields import RESULT_PATH, Field, QueryLanguage
from ..node import Node


class ResultPathMixin(Node):
    @property
    def optional_fields(self) -> list[Field]:
        fields = super().optional_fields
        if self.query_language == QueryLanguage.JSONata:
            return fields
        return [*fields, RESULT_PATH]



from ..fields import DEFAULT, END, NEXT, Field
from .mixins import AssignMixin, ChoicesMixin, OutputMixin
from .state import State


class ChoiceState(ChoicesMixin, OutputMixin, AssignMixin, State):
    @property
    def optional_fields(self) -> list[Field]:
        return [*super().optional_fields, DEFAULT]

    @property
    def forbidden_fields(self) -> list[Field]:
        return [NEXT, END, *super().forbidden_fields]



from ..problem import ProblemPredicate
from .base import Field
from .field_value import FieldValue


class AnyField(Field):
    def validate(self, value: FieldValue) -> list[ProblemPredicate]:
        return []
